//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Device/InstrumentCollection.cpp
//! @brief     Implement class InstrumentCollection
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Device/InstrumentCollection.h"
#include "GUI/Model/CatDevice/InstrumentItemCatalog.h"
#include "GUI/Model/Data/RealItem.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/Support/XML/Serialize.h"
#include "GUI/Util/String.h"

InstrumentCollection::~InstrumentCollection()
{
    qDeleteAll(m_instruments);
}

void InstrumentCollection::clear()
{
    qDeleteAll(m_instruments);
    m_instruments.clear();
}

void InstrumentCollection::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwCatalogized<InstrumentItemCatalog>(s, "InstrumentCollection", m_instruments);
}

InstrumentItem* InstrumentCollection::insertCopy(const InstrumentItem& source)
{
    auto* copy = source.createCopy();
    copy->setId(QUuid::createUuid().toString());
    m_instruments << copy;
    return copy;
}

QVector<InstrumentItem*> InstrumentCollection::collectedItems() const
{
    return m_instruments;
}

QStringList InstrumentCollection::instrumentNames() const
{
    QStringList existingNames;
    for (const auto* item : collectedItems())
        existingNames << item->instrumentName();
    return existingNames;
}

QString InstrumentCollection::suggestInstrumentName(const QString& baseName) const
{
    return GUI::Util::String::suggestName(instrumentNames(), baseName);
}

QVector<InstrumentItem*> InstrumentCollection::instrumentItems(
    const std::function<bool(const InstrumentItem*)>& accept) const
{
    QVector<InstrumentItem*> result;
    for (auto* p : collectedItems())
        if (accept(p))
            result << p;

    return result;
}

QVector<Instrument2DItem*> InstrumentCollection::instrument2DItems() const
{
    QVector<Instrument2DItem*> result;
    for (auto* p : collectedItems())
        if (auto* p2D = dynamic_cast<Instrument2DItem*>(p))
            result << p2D;

    return result;
}

InstrumentItem* InstrumentCollection::findInstrumentById(const QString& instrumentId) const
{
    for (auto* instrument : collectedItems())
        if (instrument->id() == instrumentId)
            return instrument;

    return nullptr;
}

bool InstrumentCollection::instrumentExists(const QString& instrumentId) const
{
    return findInstrumentById(instrumentId) != nullptr;
}

void InstrumentCollection::removeInstrument(InstrumentItem* instrument)
{
    m_instruments.removeAll(instrument);
    delete instrument;
}
