//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Device/MaskItems.cpp
//! @brief     Implements MaskItems classes
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Device/MaskItems.h"
#include "Device/Mask/Ellipse.h"
#include "Device/Mask/InfinitePlane.h"
#include "Device/Mask/Line.h"
#include "Device/Mask/Polygon.h"
#include "Device/Mask/Rectangle.h"
#include "GUI/Model/CatDevice/MaskItemCatalog.h"
#include "GUI/Model/Model/SessionModel.h"
#include "GUI/Support/XML/Serialize.h"

MaskContainerItem::MaskContainerItem()
    : MaskItemObject(M_TYPE)
{
    const QString T_MASKS = "Mask Tag";
    static const QStringList allowedMasks = {RectangleItem::M_TYPE,       PolygonItem::M_TYPE,
                                             EllipseItem::M_TYPE,         VerticalLineItem::M_TYPE,
                                             HorizontalLineItem::M_TYPE,  MaskAllItem::M_TYPE,
                                             RegionOfInterestItem::M_TYPE};
    registerTag(T_MASKS, 0, -1, allowedMasks);
    setDefaultTag(T_MASKS);
}

QVector<MaskItem*> MaskContainerItem::maskItems() const
{
    QVector<MaskItem*> result;
    for (auto* child : itemChildren())
        if (auto* mask = dynamic_cast<MaskItem*>(child))
            result.append(mask);

    return result;
}

void MaskContainerItem::insertMask(int row, MaskItem* maskItem)
{
    insertChild(row, maskItem);
}

void MaskContainerItem::addMask(MaskItem* maskItem)
{
    insertChild(-1, maskItem);
}

RegionOfInterestItem* MaskContainerItem::regionOfInterestItem() const
{
    return firstChildOfType<RegionOfInterestItem>();
}

void MaskContainerItem::clear()
{
    model()->removeRows(0, numberOfChildren(), index());
}

QString MaskItem::maskName() const
{
    return itemName();
}

void MaskItem::setMaskName(const QString& name)
{
    setItemName(name);
}

/* ------------------------------------------------------------------------- */

bool MaskItem::maskValue() const
{
    return getItemValue(P_MASK_VALUE).toBool();
}

void MaskItem::setMaskValue(bool mask_value)
{
    setItemValue(P_MASK_VALUE, mask_value);
    emit maskVisibilityChanged();
}

bool MaskItem::isVisibleValue() const
{
    return getItemValue(P_IS_VISIBLE).toBool();
}

void MaskItem::setIsVisibleValue(bool visible)
{
    setItemValue(P_IS_VISIBLE, visible);
    emit maskVisibilityChanged();
}

MaskItem::MaskItem(const QString& name)
    : MaskItemObject(name)
{
    addProperty(P_MASK_VALUE, true);
    addProperty(P_IS_VISIBLE, true);
}

SessionItem* MaskItem::maskValueItem() const
{
    return getItem(P_MASK_VALUE);
}

/* ------------------------------------------------------------------------- */

RectangleItem::RectangleItem()
    : RectangleItem(M_TYPE)
{
}

RectangleItem::RectangleItem(const QString& modelType)
    : MaskItem(modelType)
{
    setItemName(modelType);
    addProperty(P_XLOW, double(0.0));
    addProperty(P_YLOW, double(0.0));
    addProperty(P_XUP, double(0.0));
    addProperty(P_YUP, double(0.0));
}

std::unique_ptr<IShape2D> RectangleItem::createShape(double scale) const
{
    double xlow = scale * xLow();
    double ylow = scale * yLow();
    double xup = scale * xUp();
    double yup = scale * yUp();
    return std::make_unique<Rectangle>(xlow, ylow, xup, yup);
}

double RectangleItem::xLow() const
{
    return getItemValue(P_XLOW).toDouble();
}

void RectangleItem::setXLow(double val)
{
    setItemValue(P_XLOW, val);
    emit maskGeometryChanged();
}

double RectangleItem::yLow() const
{
    return getItemValue(P_YLOW).toDouble();
}

void RectangleItem::setYLow(double val)
{
    setItemValue(P_YLOW, val);
    emit maskGeometryChanged();
}

double RectangleItem::xUp() const
{
    return getItemValue(P_XUP).toDouble();
}

void RectangleItem::setXUp(double val)
{
    setItemValue(P_XUP, val);
    emit maskGeometryChanged();
}

double RectangleItem::yUp() const
{
    return getItemValue(P_YUP).toDouble();
}

void RectangleItem::setYUp(double val)
{
    setItemValue(P_YUP, val);
    emit maskGeometryChanged();
}

void RectangleItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwSessionItem<bool>(s, "maskValue", getItem(P_MASK_VALUE));
    Serialize::rwSessionItem<bool>(s, "visible", getItem(P_IS_VISIBLE));
    Serialize::rwSessionItem<double>(s, "xlow", getItem(P_XLOW));
    Serialize::rwSessionItem<double>(s, "ylow", getItem(P_YLOW));
    Serialize::rwSessionItem<double>(s, "xup", getItem(P_XUP));
    Serialize::rwSessionItem<double>(s, "yup", getItem(P_YUP));
}

/* ------------------------------------------------------------------------- */

RegionOfInterestItem::RegionOfInterestItem()
    : RectangleItem(M_TYPE)
{
    setMaskValue(false);
}

std::unique_ptr<IShape2D> RegionOfInterestItem::createShape(double scale) const
{
    auto shape = RectangleItem::createShape(scale);
    dynamic_cast<Rectangle*>(shape.get())->setInverted(true);
    return shape;
}

/* ------------------------------------------------------------------------- */

PolygonPointItem::PolygonPointItem()
    : MaskItemObject(M_TYPE)
{
    setItemName(M_TYPE);
    addProperty(P_POSX, double(0.0));
    addProperty(P_POSY, double(0.0));
}

double PolygonPointItem::posX() const
{
    return getItemValue(P_POSX).toDouble();
}

void PolygonPointItem::setPosX(double val)
{
    setItemValue(P_POSX, val);
    emit maskGeometryChanged();
}

double PolygonPointItem::posY() const
{
    return getItemValue(P_POSY).toDouble();
}

void PolygonPointItem::setPosY(double val)
{
    setItemValue(P_POSY, val);
    emit maskGeometryChanged();
}

void PolygonPointItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwSessionItem<double>(s, "x", getItem(P_POSX));
    Serialize::rwSessionItem<double>(s, "y", getItem(P_POSY));
}

/* ------------------------------------------------------------------------- */

PolygonItem::PolygonItem()
    : MaskItem(M_TYPE)
{
    setItemName(M_TYPE);
    const QString T_POINTS = "Point tag";
    registerTag(T_POINTS, 0, -1, {PolygonPointItem::M_TYPE});
    setDefaultTag(T_POINTS);
    addProperty(P_ISCLOSED, false);
}

std::unique_ptr<IShape2D> PolygonItem::createShape(double scale) const
{
    std::vector<double> x, y;
    for (auto* item : this->childrenOfType<PolygonPointItem>()) {
        x.push_back(scale * item->posX());
        y.push_back(scale * item->posY());
    }
    return std::make_unique<Polygon>(x, y);
}

bool PolygonItem::isClosed() const
{
    return getItemValue(P_ISCLOSED).toBool();
}

void PolygonItem::setIsClosed(bool closed)
{
    setItemValue(P_ISCLOSED, closed);
}

QVector<PolygonPointItem*> PolygonItem::points() const
{
    return childrenOfType<PolygonPointItem>();
}

void PolygonItem::addPoint(double x, double y)
{
    auto* pointItem = new PolygonPointItem();
    pointItem->setPosX(x);
    pointItem->setPosY(y);
    insertChild(-1, pointItem);
}

void PolygonItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwSessionItem<bool>(s, "maskValue", getItem(P_MASK_VALUE));
    Serialize::rwSessionItem<bool>(s, "visible", getItem(P_IS_VISIBLE));
    Serialize::rwSessionItem<bool>(s, "closed", getItem(P_ISCLOSED));

    if (s.xmlReader()) {
        QVector<PolygonPointItem*> pts;
        Serialize::rwVector(s, "points", pts);
        for (const auto* point : pts)
            addPoint(point->posX(), point->posY());
        qDeleteAll(pts);
    } else {
        auto p = points();
        Serialize::rwVector(s, "points", p);
    }
}

/* ------------------------------------------------------------------------- */

VerticalLineItem::VerticalLineItem()
    : MaskItem(M_TYPE)
{
    setItemName(M_TYPE);
    addProperty(P_POSX, double(0.0));
}

std::unique_ptr<IShape2D> VerticalLineItem::createShape(double scale) const
{
    return std::make_unique<VerticalLine>(scale * posX());
}

double VerticalLineItem::posX() const
{
    return getItemValue(P_POSX).toDouble();
}

void VerticalLineItem::setPosX(double val)
{
    setItemValue(P_POSX, val);
    emit maskGeometryChanged(this);
}

void VerticalLineItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwSessionItem<bool>(s, "maskValue", getItem(P_MASK_VALUE));
    Serialize::rwSessionItem<bool>(s, "visible", getItem(P_IS_VISIBLE));
    Serialize::rwSessionItem<double>(s, "x", getItem(P_POSX));
}

/* ------------------------------------------------------------------------- */

HorizontalLineItem::HorizontalLineItem()
    : MaskItem(M_TYPE)
{
    setItemName(M_TYPE);
    addProperty(P_POSY, double(0.0));
}

std::unique_ptr<IShape2D> HorizontalLineItem::createShape(double scale) const
{
    return std::make_unique<HorizontalLine>(scale * posY());
}

double HorizontalLineItem::posY() const
{
    return getItemValue(P_POSY).toDouble();
}

void HorizontalLineItem::setPosY(double pos_y)
{
    setItemValue(P_POSY, pos_y);
    emit maskGeometryChanged(this);
}

void HorizontalLineItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwSessionItem<bool>(s, "maskValue", getItem(P_MASK_VALUE));
    Serialize::rwSessionItem<bool>(s, "visible", getItem(P_IS_VISIBLE));
    Serialize::rwSessionItem<double>(s, "y", getItem(P_POSY));
}

/* ------------------------------------------------------------------------- */

EllipseItem::EllipseItem()
    : MaskItem(M_TYPE)
{
    setItemName(M_TYPE);
    addProperty(P_XCENTER, double(0.0));
    addProperty(P_YCENTER, double(0.0));
    addProperty(P_XRADIUS, double(0.0));
    addProperty(P_YRADIUS, double(0.0));
    addProperty(P_ANGLE, double(0.0));
}

std::unique_ptr<IShape2D> EllipseItem::createShape(double scale) const
{
    double xcenter = scale * xCenter();
    double ycenter = scale * yCenter();
    double xradius = scale * xRadius();
    double yradius = scale * yRadius();
    double _angle = scale * angle();

    return std::make_unique<Ellipse>(xcenter, ycenter, xradius, yradius, _angle);
}

double EllipseItem::xCenter() const
{
    return getItemValue(P_XCENTER).toDouble();
}

void EllipseItem::setXCenter(double val)
{
    setItemValue(P_XCENTER, val);
    emit maskGeometryChanged();
}

double EllipseItem::yCenter() const
{
    return getItemValue(P_YCENTER).toDouble();
}

void EllipseItem::setYCenter(double val)
{
    setItemValue(P_YCENTER, val);
    emit maskGeometryChanged();
}

double EllipseItem::xRadius() const
{
    return getItemValue(P_XRADIUS).toDouble();
}

void EllipseItem::setXRadius(double val)
{
    setItemValue(P_XRADIUS, val);
    emit maskGeometryChanged();
}

double EllipseItem::yRadius() const
{
    return getItemValue(P_YRADIUS).toDouble();
}

void EllipseItem::setYRadius(const double val)
{
    setItemValue(P_YRADIUS, val);
    emit maskGeometryChanged();
}

double EllipseItem::angle() const
{
    return getItemValue(P_ANGLE).toDouble();
}

void EllipseItem::setAngle(const double val)
{
    setItemValue(P_ANGLE, val);
    emit maskGeometryChanged();
}

void EllipseItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwSessionItem<bool>(s, "maskValue", getItem(P_MASK_VALUE));
    Serialize::rwSessionItem<bool>(s, "visible", getItem(P_IS_VISIBLE));
    Serialize::rwSessionItem<double>(s, "x", getItem(P_XCENTER));
    Serialize::rwSessionItem<double>(s, "y", getItem(P_YCENTER));
    Serialize::rwSessionItem<double>(s, "xr", getItem(P_XRADIUS));
    Serialize::rwSessionItem<double>(s, "yr", getItem(P_YRADIUS));
    Serialize::rwSessionItem<double>(s, "angle", getItem(P_ANGLE));
}

/* ------------------------------------------------------------------------- */

MaskAllItem::MaskAllItem()
    : MaskItem(M_TYPE)
{
    setItemName(M_TYPE);
    maskValueItem()->setEnabled(false);
}

std::unique_ptr<IShape2D> MaskAllItem::createShape(double) const
{
    return std::make_unique<InfinitePlane>();
}

void MaskAllItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwSessionItem<bool>(s, "maskValue", getItem(P_MASK_VALUE));
    Serialize::rwSessionItem<bool>(s, "visible", getItem(P_IS_VISIBLE));
}

/* ------------------------------------------------------------------------- */

MaskItems::MaskItems()
    : SessionModel("MaskItems")
{
    m_maskContainer = insertItem<MaskContainerItem>();
}

void MaskItems::insertMask(int index, MaskItem* maskItem)
{
    m_maskContainer->insertChild(index, maskItem);
}

QVector<MaskItem*> MaskItems::maskItems() const
{
    return m_maskContainer->maskItems();
}

void MaskItems::copy(const MaskContainerItem* maskContainer)
{
    m_maskContainer->clear();
    if (maskContainer)
        for (auto mask : maskContainer->maskItems())
            copyItem(mask, m_maskContainer);
}

void MaskItems::clear()
{
    SessionModel::clear();
    m_maskContainer = insertItem<MaskContainerItem>();
}

void MaskItems::serialize(Streamer& s)
{
    QVector<MaskItem*> items;
    if (!s.xmlReader())
        items = maskItems();

    s.assertVersion(0);
    Serialize::rwCatalogized<MaskItemCatalog>(s, "maskItems", items);

    if (s.xmlReader())
        for (auto m : items)
            m_maskContainer->addMask(m);
}

MaskItemObject::MaskItemObject(const QString& modelType)
    : SessionItem(modelType)
{
}

MaskItemObject::~MaskItemObject()
{
    emit maskToBeDestroyed(this);
}
