//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Support/XML/Serialize.cpp
//! @brief     Implements class Streamer
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Support/XML/Serialize.h"
#include "GUI/Support/XML/UtilXML.h"
#include <QColor>

void Serialize::rwValue(Streamer& s, const QString& tag, bool& val)
{
    if (QXmlStreamWriter* w = s.xmlWriter()) {
        w->writeStartElement(tag);
        w->writeAttribute(StreamerXML::Tags::Value, val ? "1" : "0");
        w->writeEndElement();
    } else if (QXmlStreamReader* r = s.xmlReader()) {
        r->readNextStartElement();
        s.assertCurrentTag(tag);
        val = r->attributes().value(StreamerXML::Tags::Value).toUInt() > 0;
        s.gotoEndElementOfTag(tag);
    } else
        ASSERT(0);
}

void Serialize::rwValue(Streamer& s, const QString& tag, int& val)
{
    if (QXmlStreamWriter* w = s.xmlWriter()) {
        w->writeStartElement(tag);
        w->writeAttribute(StreamerXML::Tags::Value, QString::number(val));
        w->writeEndElement();
    } else if (QXmlStreamReader* r = s.xmlReader()) {
        r->readNextStartElement();
        s.assertCurrentTag(tag);
        val = r->attributes().value(StreamerXML::Tags::Value).toInt();
        s.gotoEndElementOfTag(tag);
    } else
        ASSERT(0);
}

void Serialize::rwValue(Streamer& s, const QString& tag, uint& val)
{
    if (QXmlStreamWriter* w = s.xmlWriter()) {
        w->writeStartElement(tag);
        w->writeAttribute(StreamerXML::Tags::Value, QString::number(val));
        w->writeEndElement();
    } else if (QXmlStreamReader* r = s.xmlReader()) {
        r->readNextStartElement();
        s.assertCurrentTag(tag);
        val = r->attributes().value(StreamerXML::Tags::Value).toInt();
        s.gotoEndElementOfTag(tag);
    } else
        ASSERT(0);
}

void Serialize::rwValue(Streamer& s, const QString& tag, double& val)
{
    if (QXmlStreamWriter* w = s.xmlWriter()) {
        w->writeStartElement(tag);
        XML::writeAttribute(w, StreamerXML::Tags::Value, val);
        w->writeEndElement();
    } else if (QXmlStreamReader* r = s.xmlReader()) {
        r->readNextStartElement();
        s.assertCurrentTag(tag);
        val = r->attributes().value(StreamerXML::Tags::Value).toDouble();
        s.gotoEndElementOfTag(tag);
    } else
        ASSERT(0);
}

void Serialize::rwValue(Streamer& s, const QString& tag, QString& val)
{
    if (QXmlStreamWriter* w = s.xmlWriter()) {
        w->writeStartElement(tag);
        w->writeAttribute(StreamerXML::Tags::Value, val);
        w->writeEndElement();
    } else if (QXmlStreamReader* r = s.xmlReader()) {
        r->readNextStartElement();
        s.assertCurrentTag(tag);
        XML::readAttribute(r, StreamerXML::Tags::Value, &val);
        s.gotoEndElementOfTag(tag);
    } else
        ASSERT(0);
}

void Serialize::rwValue(Streamer& s, const QString& tag, QColor& col)
{
    if (QXmlStreamWriter* w = s.xmlWriter()) {
        w->writeStartElement(tag);
        w->writeAttribute(StreamerXML::Tags::Value, col.isValid() ? col.name(QColor::HexArgb) : "");
        w->writeEndElement();
    } else if (QXmlStreamReader* r = s.xmlReader()) {
        r->readNextStartElement();
        s.assertCurrentTag(tag);
        QString colName = r->attributes().value(StreamerXML::Tags::Value).toString();
        col = QColor(colName);
        s.gotoEndElementOfTag(tag);
    } else
        ASSERT(0);
}

void Serialize::rwValue(Streamer& s, const QString& tag, QByteArray& val)
{
    if (QXmlStreamWriter* w = s.xmlWriter()) {
        w->writeStartElement(tag);
        w->writeCharacters(val.toBase64());
        w->writeEndElement();
    } else if (QXmlStreamReader* r = s.xmlReader()) {
        r->readNextStartElement();
        s.assertCurrentTag(tag);
        QString valueAsBase64 = r->readElementText(QXmlStreamReader::SkipChildElements);
        val = QByteArray::fromBase64(valueAsBase64.toLatin1());
        s.gotoEndElementOfTag(tag);
    } else
        ASSERT(0);
}

void Serialize::rwValue(Streamer& s, const QString& tag, R3& val)
{
    if (QXmlStreamWriter* w = s.xmlWriter()) {
        w->writeStartElement(tag);
        XML::writeAttribute(w, StreamerXML::Tags::Value, val);
        w->writeEndElement();
    } else if (QXmlStreamReader* r = s.xmlReader()) {
        s.gotoStartElementOfTag(tag);
        XML::readAttribute(r, StreamerXML::Tags::Value, &val);
        s.gotoEndElementOfTag(tag);
    } else
        ASSERT(0);
}
