//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Fit/MinimizerSettingsWidget.cpp
//! @brief     Implements class MinimizerSettingsWidget
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Fit/MinimizerSettingsWidget.h"
#include "GUI/Model/Job/FitSuiteItem.h"
#include "GUI/Model/Job/JobItem.h"
#include "GUI/View/Common/DoubleSpinBox.h"
#include "GUI/View/Common/SafeSpinBox.h"
#include "GUI/View/Tool/LayoutUtils.h"
#include "GUI/View/Tool/WidgetUtils.h"
#include <QComboBox>
#include <QFormLayout>
#include <QPushButton>
#include <QStandardItemModel>
#include <QVBoxLayout>

MinimizerSettingsWidget::MinimizerSettingsWidget(QWidget* parent)
    : QWidget(parent)
    , m_currentItem(nullptr)
{
    setWindowTitle(QLatin1String("Minimizer Settings"));
    setAttribute(Qt::WA_StyledBackground, true);
    setProperty("stylable", true); // for stylesheet addressing

    m_mainLayout = new QFormLayout(this);
    m_mainLayout->setContentsMargins(8, 8, 8, 8);
    m_mainLayout->setSpacing(5);
}

void MinimizerSettingsWidget::setJobItem(JobItem* jobItem)
{
    ASSERT(jobItem);
    setMinItem(jobItem->fitSuiteItem()->minimizerContainerItem());
}

void MinimizerSettingsWidget::setMinItem(MinimizerContainerItem* minimizerItem)
{
    ASSERT(minimizerItem);

    GUI::Util::Layout::clearLayout(m_mainLayout);
    m_updaters.clear();
    m_currentItem = minimizerItem;

    if (!m_currentItem)
        return;

    createGroupedAlgorithmsCombo();

    auto* w = new QWidget(this);
    m_minimizerLayout = new QFormLayout(w);
    m_minimizerLayout->setContentsMargins(10, 8, 0, 8);
    m_mainLayout->addRow(w);

    m_mainLayout->addRow("Objective metric:",
                         GUI::Util::createComboBox(
                             [=] { return m_currentItem->objectiveMetricCombo(); },
                             [=](const QString& t) { m_currentItem->setCurrentObjectiveMetric(t); },
                             &m_updaters,
                             "Objective metric to use for estimating distance between simulated "
                             "and experimental data"));
    m_mainLayout->addRow("Norm function:",
                         GUI::Util::createComboBox(
                             [=] { return m_currentItem->normFunctionCombo(); },
                             [=](const QString& t) { m_currentItem->setCurrentNormFunction(t); },
                             &m_updaters,
                             "Normalization to use for estimating distance between simulated and "
                             "experimental data"));

    createMimimizerEdits();
    updateUIValues();
}

void MinimizerSettingsWidget::createGroupedAlgorithmsCombo()
{
    QComboBox* comboBox = new QComboBox;
    QStringList list = m_currentItem->commonAlgorithmCombo().values();

    // list with headers and separators
    QList<qsizetype> header_indices;
    QList<qsizetype> non_separator_indices;
    for (QString algorithm : list) {
        comboBox->addItem(algorithm);
        if (!m_currentItem->algorithmHasMinimizer(algorithm)) {

            comboBox->insertSeparator(comboBox->count() - 1);

            qsizetype header_index = comboBox->count() - 1;
            header_indices.append(header_index);
            non_separator_indices.append(header_index);
            QStandardItemModel* model = qobject_cast<QStandardItemModel*>(comboBox->model());
            QStandardItem* header_item = model->item(header_index);
            header_item->setSelectable(false);

            QFont font(comboBox->font());
            font.setBold(true);
            header_item->setFont(font);

            comboBox->insertSeparator(comboBox->count());
        } else
            non_separator_indices.append(comboBox->count() - 1);
    }
    comboBox->setCurrentText(m_currentItem->commonAlgorithmCombo().currentValue());

    // tooltips
    QStringList tooltips = m_currentItem->commonAlgorithmCombo().toolTips();
    ASSERT(tooltips.size() == list.size())
    int list_index = 0;
    for (int index : non_separator_indices)
        comboBox->setItemData(index, tooltips.at(list_index++), Qt::ToolTipRole);

    // action
    comboBox->setProperty("previous", comboBox->currentIndex());
    QObject::connect(comboBox, &QComboBox::currentTextChanged, [=] {
        // skip headers while scrolling
        if (header_indices.contains(comboBox->currentIndex())) {
            int previous_state = comboBox->property("previous").toInt();

            int prev = comboBox->currentIndex() - 2;
            int next = comboBox->currentIndex() + 2;

            QSignalBlocker b(comboBox);
            if (previous_state < comboBox->currentIndex() && next < comboBox->count())
                comboBox->setCurrentIndex(next);
            else if (previous_state > comboBox->currentIndex() && prev >= 0)
                comboBox->setCurrentIndex(prev);
            else
                comboBox->setCurrentIndex(previous_state);
        }
        comboBox->setProperty("previous", comboBox->currentIndex());

        m_currentItem->setCurrentCommonAlgorithm(comboBox->currentText());
        createMimimizerEdits();
    });

    // update state
    m_updaters << [=]() {
        QSignalBlocker b(comboBox);
        comboBox->setCurrentText(m_currentItem->commonAlgorithmCombo().currentValue());
    };

    m_mainLayout->addRow("Algorithm:", comboBox);
}

void MinimizerSettingsWidget::createMimimizerEdits()
{
    GUI::Util::Layout::clearLayout(m_minimizerLayout);

    // Minuit2
    if (m_currentItem->currentMinimizer() == minimizerTypeToName(MinimizerType::Minuit2))
        createMinuitEdits();

    // GSL MultiMin
    if (m_currentItem->currentMinimizer() == minimizerTypeToName(MinimizerType::GSLMultiMin))
        createGSLMultiMinEdits();

    // TMVA Genetic
    if (m_currentItem->currentMinimizer() == minimizerTypeToName(MinimizerType::Genetic))
        createTMVAGeneticEdits();

    // GSL Simulated Annealing
    if (m_currentItem->currentMinimizer() == minimizerTypeToName(MinimizerType::GSLSimAn))
        createGSLSimulatedAnnealingEdits();

    // GSL Levenberg-Marquardt
    if (m_currentItem->currentMinimizer() == minimizerTypeToName(MinimizerType::GSLLMA))
        createGSLLevMarEdits();
}

void MinimizerSettingsWidget::createMinuitEdits()
{
    MinuitMinimizerItem* minItem = m_currentItem->minimizerItemMinuit();

    m_minimizerLayout->addRow(
        "Strategy:",
        GUI::Util::createIntSpinbox(
            [=] { return minItem->strategy(); }, [=](int v) { minItem->setStrategy(v); },
            RealLimits::limited(0, 2), "Minimization strategy (0-low, 1-medium, 2-high quality)",
            &m_updaters, true /*easy scroll*/));

    m_minimizerLayout->addRow("ErrorDef factor:",
                              GUI::Util::createDoubleSpinbox(
                                  [=] { return minItem->errorDefinition(); },
                                  [=](double v) { minItem->setErrorDefinition(v); }, &m_updaters,
                                  "Error definition factor for parameter error calculation",
                                  RealLimits::positive()));

    m_minimizerLayout->addRow(
        "Tolerance:",
        GUI::Util::createDoubleSpinbox([=] { return minItem->tolerance(); },
                                       [=](double v) { minItem->setTolerance(v); }, &m_updaters,
                                       "Tolerance on the function value at the minimum",
                                       RealLimits::nonnegative()));

    m_minimizerLayout->addRow(
        "Precision:",
        GUI::Util::createDoubleSpinbox([=] { return minItem->precision(); },
                                       [=](double v) { minItem->setPrecision(v); }, &m_updaters,
                                       "Relative floating point arithmetic precision",
                                       RealLimits::nonnegative()));

    m_minimizerLayout->addRow(
        "Max func calls:",
        GUI::Util::createIntSpinbox([=] { return minItem->maxFuncCalls(); },
                                    [=](int v) { minItem->setMaxFuncCalls(v); },
                                    RealLimits::nonnegative(), "Maximum number of function calls",
                                    &m_updaters, true /*easy scroll*/));
}

void MinimizerSettingsWidget::createGSLMultiMinEdits()
{
    GSLMultiMinimizerItem* minItem = m_currentItem->minimizerItemGSLMulti();

    m_minimizerLayout->addRow(
        "Max iterations:",
        GUI::Util::createIntSpinbox([=] { return minItem->maxIterations(); },
                                    [=](int v) { minItem->setMaxIterations(v); },
                                    RealLimits::nonnegative(), "Maximum number of iterations",
                                    &m_updaters, true /*easy scroll*/));
}

void MinimizerSettingsWidget::createTMVAGeneticEdits()
{
    GeneticMinimizerItem* minItem = m_currentItem->minimizerItemGenetic();

    m_minimizerLayout->addRow(
        "Tolerance:",
        GUI::Util::createDoubleSpinbox([=] { return minItem->tolerance(); },
                                       [=](double v) { minItem->setTolerance(v); }, &m_updaters,
                                       "Tolerance on the function value at the minimum",
                                       RealLimits::nonnegative()));

    m_minimizerLayout->addRow(
        "Max iterations:",
        GUI::Util::createIntSpinbox([=] { return minItem->maxIterations(); },
                                    [=](int v) { minItem->setMaxIterations(v); },
                                    RealLimits::nonnegative(), "Maximum number of iterations",
                                    &m_updaters, true /*easy scroll*/));

    m_minimizerLayout->addRow(
        "Population:", GUI::Util::createIntSpinbox([=] { return minItem->populationSize(); },
                                                   [=](int v) { minItem->setPopulationSize(v); },
                                                   RealLimits::nonnegative(), "Population size",
                                                   &m_updaters, true /*easy scroll*/));

    m_minimizerLayout->addRow("Random seed:", GUI::Util::createIntSpinbox(
                                                  [=] { return minItem->randomSeed(); },
                                                  [=](int v) { minItem->setRandomSeed(v); },
                                                  RealLimits::limitless(),
                                                  "Initialization of pseudorandom number generator",
                                                  &m_updaters, true /*easy scroll*/));
}

void MinimizerSettingsWidget::createGSLSimulatedAnnealingEdits()
{
    SimAnMinimizerItem* minItem = m_currentItem->minimizerItemSimAn();

    m_minimizerLayout->addRow(
        "Max iterations:", GUI::Util::createIntSpinbox([=] { return minItem->maxIterations(); },
                                                       [=](int v) { minItem->setMaxIterations(v); },
                                                       RealLimits::nonnegative(),
                                                       "Number of points to try for each step",
                                                       &m_updaters, true /*easy scroll*/));

    m_minimizerLayout->addRow(
        "Iterations at T:",
        GUI::Util::createIntSpinbox(
            [=] { return minItem->iterationsAtEachTemp(); },
            [=](int v) { minItem->setIterationsAtEachTemp(v); }, RealLimits::nonnegative(),
            "Number of iterations at each temperature", &m_updaters, true /*easy scroll*/));

    m_minimizerLayout->addRow("Step size:",
                              GUI::Util::createDoubleSpinbox(
                                  [=] { return minItem->stepSize(); },
                                  [=](double v) { minItem->setStepSize(v); }, &m_updaters,
                                  "Max step size used in random walk", RealLimits::nonnegative()));

    m_minimizerLayout->addRow("k:", GUI::Util::createDoubleSpinbox(
                                        [=] { return minItem->boltzmanK(); },
                                        [=](double v) { minItem->setBoltzmanK(v); }, &m_updaters,
                                        "Boltzmann k", RealLimits::nonnegative()));

    m_minimizerLayout->addRow(
        "T init:",
        GUI::Util::createDoubleSpinbox([=] { return minItem->boltzmanInitT(); },
                                       [=](double v) { minItem->setBoltzmanInitT(v); }, &m_updaters,
                                       "Boltzmann initial temperature", RealLimits::nonnegative()));

    m_minimizerLayout->addRow("mu:", GUI::Util::createDoubleSpinbox(
                                         [=] { return minItem->boltzmanMu(); },
                                         [=](double v) { minItem->setBoltzmanMu(v); }, &m_updaters,
                                         "Boltzmann mu", RealLimits::nonnegative()));

    m_minimizerLayout->addRow(
        "T min:",
        GUI::Util::createDoubleSpinbox([=] { return minItem->boltzmanMinT(); },
                                       [=](double v) { minItem->setBoltzmanMinT(v); }, &m_updaters,
                                       "Boltzmann minimal temperature", RealLimits::nonnegative()));
}

void MinimizerSettingsWidget::createGSLLevMarEdits()
{
    GSLLMAMinimizerItem* minItem = m_currentItem->minimizerItemGSLLMA();

    m_minimizerLayout->addRow(
        "Tolerance:",
        GUI::Util::createDoubleSpinbox([=] { return minItem->tolerance(); },
                                       [=](double v) { minItem->setTolerance(v); }, &m_updaters,
                                       "Tolerance on the function value at the minimum",
                                       RealLimits::nonnegative()));

    m_minimizerLayout->addRow(
        "Max iterations:",
        GUI::Util::createIntSpinbox([=] { return minItem->maxIterations(); },
                                    [=](int v) { minItem->setMaxIterations(v); },
                                    RealLimits::nonnegative(), "Maximum number of iterations",
                                    &m_updaters, true /*easy scroll*/));
}

void MinimizerSettingsWidget::updateUIValues()
{
    for (const auto& updater : m_updaters)
        updater();
}
