//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sim/Contrib/ParticleLayoutContribution.cpp
//! @brief     Implements class ParticleLayoutContribution.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Sim/Contrib/ParticleLayoutContribution.h"
#include "Resample/Element/DiffuseElement.h"
#include "Resample/Interparticle/DecouplingApproximationStrategy.h"
#include "Resample/Interparticle/SSCAStrategy.h"
#include "Resample/Processed/ReLayout.h"
#include "Sample/Aggregate/InterferenceRadialParacrystal.h"

namespace {

std::unique_ptr<IInterparticleStrategy> processedInterference(const ReLayout& re_layout,
                                                              const SimulationOptions& sim_params,
                                                              bool polarized)
{
    const IInterference* iff = re_layout.interferenceFunction();

    const OwningVector<const CoheringSubparticles>& weighted_formfactors = re_layout.subparticles();

    const auto* radial_para = dynamic_cast<const InterferenceRadialParacrystal*>(iff);
    if (radial_para)
        if (double kappa = radial_para->kappa(); kappa > 0.0)
            return std::make_unique<SSCAStrategy>(weighted_formfactors, radial_para, sim_params,
                                                  polarized, kappa);

    return std::make_unique<DecouplingApproximationStrategy>(weighted_formfactors, iff, sim_params,
                                                             polarized);
}

} // namespace


ParticleLayoutContribution::ParticleLayoutContribution(const ReLayout& layout,
                                                       const SimulationOptions& options,
                                                       bool polarized)
    : m_layout(layout)
    , m_interparticle_strategy(processedInterference(layout, options, polarized))
{
}

ParticleLayoutContribution::~ParticleLayoutContribution() = default;

void ParticleLayoutContribution::compute(DiffuseElement& ele) const
{
    ele.addIntensity(m_interparticle_strategy->evaluate(ele) * m_layout.surfaceDensity());
}
