#include "GUI/Model/BaseItem/PropertyItem.h"
#include "GUI/Model/Data/DataItem.h"
#include "GUI/Model/Data/RealItem.h"
#include "GUI/Model/Device/InstrumentCollection.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/Model/Device/MaskItems.h"
#include "GUI/Model/Job/JobItem.h"
#include "GUI/Model/Model/JobModel.h"
#include "GUI/Model/Sample/LayerItem.h"
#include "GUI/Model/Sample/SampleItem.h"
#include "GUI/Support/Data/SessionItemTags.h"
#include "Tests/GTestWrapper/google_test.h"
#include <QSignalSpy>
#include <QXmlStreamWriter>
#include <memory>

class TestSessionModel : public ::testing::Test {};

//! Testing SessionModel::setData notifications.

TEST_F(TestSessionModel, setData)
{
    SessionModel model("TestModel");
    auto* item = model.insertItem<PropertyItem>();

    QSignalSpy spy(&model, &SessionModel::dataChanged);

    QVariant data(42.0);

    // setting the data and checking that signal was emmitted only once
    EXPECT_TRUE(model.setData(model.indexOfItem(item), data, Qt::DisplayRole));
    EXPECT_EQ(spy.count(), 1);
    EXPECT_EQ(item->value().toDouble(), 42.0);
    QList<QVariant> arguments = spy.takeFirst();
    EXPECT_EQ(arguments.at(0).toModelIndex(), model.indexOfItem(item));

    // setting the same data second time
    EXPECT_FALSE(model.setData(model.indexOfItem(item), data, Qt::DisplayRole));
    EXPECT_EQ(spy.count(), 0);

    // setting another data
    EXPECT_TRUE(model.setData(model.indexOfItem(item), QVariant(43.0), Qt::DisplayRole));
    EXPECT_EQ(spy.count(), 1);
}

//! Test if SessionItem can be copied from one model to another. Particularly, we test
//! here if a SampleItem can be copied from SampleModel to the JobItem of JobModel

TEST_F(TestSessionModel, copyItem)
{
    auto sample1 = std::make_unique<SampleItem>();
    sample1->setSampleName("sample1");
    sample1->addStandardMaterials();

    InstrumentCollection instrumentItems;
    auto* instrument1 = instrumentItems.addInstrument<GISASInstrumentItem>();
    instrument1->setInstrumentName("instrument1");

    JobModel jobModel;
    auto* jobItem = jobModel.insertItem<JobItem>();

    jobItem->copySampleIntoJob(sample1.get());
    EXPECT_NE(jobItem->sampleItem(), nullptr);

    jobItem->copyInstrumentIntoJob(instrument1);
    EXPECT_NE(jobItem->instrumentItem(), nullptr);
}

TEST_F(TestSessionModel, moveItemFromRoot)
{
    SessionModel model("TestModel");
    auto* poly = model.insertItem<PolygonItem>();
    auto* point = model.insertItem<PolygonPointItem>();

    EXPECT_EQ(poly->itemParent(), model.rootItem());
    EXPECT_EQ(point->itemParent(), model.rootItem());
    EXPECT_EQ(model.rootItem()->numberOfChildren(), 2);

    // moving from rootItem to poly
    auto* moved = model.moveItem(point, poly);
    EXPECT_EQ(model.rootItem()->numberOfChildren(), 1);
    EXPECT_EQ(moved, point);
    EXPECT_EQ(point->itemParent(), poly);
    EXPECT_EQ(poly->getItem(), point);

    // moving from poly to rootItem
    moved = model.moveItem(point, nullptr);
    EXPECT_EQ(model.rootItem()->numberOfChildren(), 2);
    EXPECT_EQ(moved, point);
    EXPECT_EQ(point->itemParent(), model.rootItem());
    EXPECT_EQ(poly->getItem(), nullptr);
}

TEST_F(TestSessionModel, moveBetweenParents)
{
    SessionModel model("TestModel");
    auto* poly1 = model.insertItem<PolygonItem>();
    auto* point11 = model.insertItem<PolygonPointItem>(poly1);
    auto* point12 = model.insertItem<PolygonPointItem>(poly1);
    auto* poly2 = model.insertItem<PolygonItem>();

    EXPECT_EQ(point11->itemParent(), poly1);
    EXPECT_EQ(point12->itemParent(), poly1);
    EXPECT_EQ(poly1->getItem(), point11);

    auto* moved = model.moveItem(point11, poly2);
    EXPECT_EQ(moved, point11);
    EXPECT_EQ(point11->itemParent(), poly2);
    EXPECT_EQ(poly1->getItem(), point12);
}

TEST_F(TestSessionModel, moveWithinSameParent)
{
    SessionModel model("TestModel");
    auto* poly = model.insertItem<PolygonItem>();
    auto* pA = model.insertItem<PolygonPointItem>(poly);
    auto* pB = model.insertItem<PolygonPointItem>(poly);
    auto* pC = model.insertItem<PolygonPointItem>(poly);
    auto* pD = model.insertItem<PolygonPointItem>(poly);
    auto* pE = model.insertItem<PolygonPointItem>(poly);

    // 0  pA -> pA
    // 1  pB -> pC
    // 2  pC -> pB
    // 3  pD -> pD
    // 4  pE -> pE

    model.moveItem(pB, poly, 3);
    EXPECT_EQ(poly->getItems().indexOf(pA), 0);
    EXPECT_EQ(poly->getItems().indexOf(pB), 2);
    EXPECT_EQ(poly->getItems().indexOf(pC), 1);
    EXPECT_EQ(poly->getItems().indexOf(pD), 3);
    EXPECT_EQ(poly->getItems().indexOf(pE), 4);
}
